@@ -1,12 +1,12 @@
 import sys
 import json
 import re
-from deriva.core import ErmrestCatalog, AttrDict, ermrest_config, get_credential, __version__ as VERSION, \
+from deriva.core import ErmrestCatalog, AttrDict, ermrest_model, get_credential, __version__ as VERSION, \
     format_exception, urlquote
-from deriva.core.ermrest_config import CatalogColumn, CatalogForeignKey
 from deriva.config.base_config import BaseSpec, BaseSpecList, ConfigUtil, ConfigBaseCLI
 from requests.exceptions import HTTPError
 from uuid import UUID
+import warnings
 
 
 class NoForeignKeyError(ValueError):
@@ -37,11 +37,14 @@ class AclConfig:
     GC_NAME = 'groups'
     ACL_TYPES = ["catalog_acl", "schema_acls", "table_acls", "column_acls", "foreign_key_acls"]
     GLOBUS_PREFIX = 'https://auth.globus.org/'
+    ROBOT_PREFIX_FORMAT = 'https://{server}/webauthn_robot/'
 
     def __init__(self, server, catalog_id, config_file, credentials, schema_name=None, table_name=None, verbose=False):
         self.config = json.load(open(config_file))
         self.ignored_schema_patterns = []
         self.verbose = verbose
+        self.server = server
+        self.catalog_id = catalog_id
         ip = self.config.get("ignored_schema_patterns")
         if ip is not None:
             for p in ip:
@@ -56,14 +59,12 @@ class AclConfig:
         self.expand_acl_definitions()
         self.acl_bindings = self.config.get("acl_bindings")
         self.invalidate_bindings = self.config.get("invalidate_bindings")
-        self.server = server
-        self.catalog_id = catalog_id
 
         old_catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
-        self.saved_toplevel_config = ConfigUtil.find_toplevel_node(old_catalog.getCatalogConfig(), schema_name,
+        self.saved_toplevel_config = ConfigUtil.find_toplevel_node(old_catalog.getCatalogModel(), schema_name,
                                                                    table_name)
         self.catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
-        self.toplevel_config = ConfigUtil.find_toplevel_node(self.catalog.getCatalogConfig(), schema_name, table_name)
+        self.toplevel_config = ConfigUtil.find_toplevel_node(self.catalog.getCatalogModel(), schema_name, table_name)
 
     def make_speclist(self, name):
         d = self.config.get(name)
@@ -86,14 +87,14 @@ class AclConfig:
             node.acl_bindings[binding_name] = self.expand_acl_binding(binding, table_node)
         except NoForeignKeyError as e:
             detail = ''
-            if isinstance(node, CatalogColumn):
+            if isinstance(node, ermrest_model.Column):
                 detail = 'on column {n}'.format(n=node.name)
-            elif isinstance(node, CatalogForeignKey):
+            elif isinstance(node, ermrest_model.ForeignKey):
                 detail = 'on foreign key {s}.{n}'.format(s=node.names[0][0], n=node.names[0][1])
             else:
                 detail = ' {t}'.format(t=type(node))
             print(
-                "couldn't expand acl binding {b} {d} table {s}.{t}".format(b=binding_name, d=detail, s=table_node.sname,
+                "couldn't expand acl binding {b} {d} table {s}.{t}".format(b=binding_name, d=detail, s=table_node.schema.name,
                                                                            t=table_node.name))
             raise e
 
@@ -142,10 +143,10 @@ class AclConfig:
         for fkey in table_node.foreign_keys:
             if len(fkey.foreign_key_columns) == 1:
                 col = fkey.foreign_key_columns[0]
-                if col.get("table_name") == table_node.name and col.get("schema_name") == table_node.sname and col.get(
+                if col.get("table_name") == table_node.name and col.get("schema_name") == table_node.schema.name and col.get(
                         "column_name") == col_name:
                     return fkey.names[0]
-        raise NoForeignKeyError("can't find foreign key for column %I.%I(%I)", table_node.sname, table_node.name,
+        raise NoForeignKeyError("can't find foreign key for column %I.%I(%I)", table_node.schema.name, table_node.name,
                                 col_name)
 
     def set_node_acl_bindings(self, node, table_node, binding_list, invalidate_list):
@@ -156,11 +157,10 @@ class AclConfig:
         if invalidate_list is not None:
             for binding_name in invalidate_list:
                 if binding_list and binding_name in binding_list:
-                    raise ValueError("Binding {b} appears in both acl_bindings and invalidate_bindings for table {s}.{t} node {n}".format(
-                        b=binding_name,s=table_node.sname, t=table_node.name, n=node.name))
+                    raise ValueError(
+                        "Binding {b} appears in both acl_bindings and invalidate_bindings for table {s}.{t} node {n}".format(
+                            b=binding_name, s=table_node.schema.name, t=table_node.name, n=node.name))
                 node.acl_bindings[binding_name] = False
-            
-                
 
     def save_groups(self):
         glt = self.create_or_validate_group_table()
@@ -208,36 +208,34 @@ class AclConfig:
         assert schema is not None
         glt = Table(schema['tables'].get(tname))
         if glt == {}:
-            glt_spec = {
-                'comment': "Named lists of groups used in ACLs. Maintained by the rbk_acls program. "
-                           "Do not update this table manually.",
-                'annotations': {
-                    'tag:isrd.isi.edu,2016:generated': None
-                },
-                'column_definitions': [
-                    {
-                        'name': self.NC_NAME,
-                        'type': {'typename': 'text'},
-                        'nullok': False,
-                        'comment': 'Name of grouplist, used in foreign keys. This table is maintained by the rbk_acls '
-                                   'program and should not be updated by hand.'
-                    },
-                    {
-                        'name': self.GC_NAME,
-                        'type': {'base_type': {'typename': 'text'}, 'is_array': True},
-                        'nullok': True,
-                        'comment': 'List of groups. This table is maintained by the rbk_acls program and should not be '
-                                   'updated by hand.'
-                    }
-
+            glt_spec = ermrest_model.Table.define(
+                tname,
+                column_defs=[
+                    ermrest_model.Column.define(
+                        self.NC_NAME,
+                        ermrest_model.builtin_types.text,
+                        nullok=False,
+                        comment='Name of grouplist, used in foreign keys. This table is maintained by the acl-config '
+                                'program and should not be updated by hand.'
+                    ),
+                    ermrest_model.Column.define(
+                        self.GC_NAME,
+                        ermrest_model.builtin_types['text[]'],
+                        nullok=True,
+                        comment='List of groups. This table is maintained by the acl-config program and should not be '
+                                'updated by hand.'
+                    )
                 ],
-                'keys': [
-                    {
-                        'names': [[sname, "{t}_{c}_u".format(t=tname, c=self.NC_NAME)]],
-                        'unique_columns': [self.NC_NAME]
-                    }
-                ]
-            }
+                key_defs=[
+                    ermrest_model.Key.define(
+                        [self.NC_NAME],
+                        constraint_names=[[sname, "{t}_{c}_u".format(t=tname, c=self.NC_NAME)]]
+                    )
+                ],
+                comment="Named lists of groups used in ACLs. Maintained by the acl-config program. Do not update this "
+                        "table manually.",
+                annotations={'tag:isrd.isi.edu,2016:generated': None}
+            )
             glt = Table(self.create_table(sname, tname, glt_spec))
 
         else:
@@ -249,14 +247,13 @@ class AclConfig:
             if name_col.get('nullok'):
                 raise ValueError(
                     "{n} column in group list table ({s}.{t}) allows nulls".format(n=self.NC_NAME, s=sname, t=tname))
+
             nc_uniq = False
             for key in glt.get('keys'):
                 cols = key.get('unique_columns')
                 if len(cols) == 1 and cols[0] == self.NC_NAME:
                     nc_uniq = True
                     break
-                if nc_uniq:
-                    break
             if not nc_uniq:
                 raise ValueError(
                     "{n} column in group list table ({s}.{t}) is not a key".format(n=self.NC_NAME, s=sname, t=tname))
@@ -292,8 +289,10 @@ class AclConfig:
             return
         elif group.startswith(self.GLOBUS_PREFIX):
             self.validate_globus_group(group)
+        elif group.startswith(self.ROBOT_PREFIX_FORMAT.format(server=self.server)):
+            self.validate_webauthn_robot(group)
         else:
-            raise ValueError("Can't determine format of group '{g}'".format(g=group))
+            warnings.warn("Can't determine format of group '{g}'".format(g=group))
 
     def validate_globus_group(self, group):
         guid = group[len(self.GLOBUS_PREFIX):]
@@ -304,6 +303,13 @@ class AclConfig:
         if self.verbose:
             print("group '{g}' appears to be a syntactically-correct Globus group".format(g=group))
 
+    def validate_webauthn_robot(self, group):
+        robot_name = group[len(self.ROBOT_PREFIX_FORMAT.format(server=self.server)):]
+        if not robot_name:
+            raise ValueError("Group '{g}' appears to be a malformed webauthn robot identity".format(g=group))
+        if self.verbose:
+            print("group '{g}' appears to be a syntactically-correct webauthn robot identity".format(g=group))
+
     def expand_group(self, group_name):
         groups = []
         for child_name in self.groups.get(group_name):
@@ -333,7 +339,7 @@ class AclConfig:
             spec[op_type] = groups
 
     def set_table_acls(self, table):
-        spec = self.acl_specs["table_acls"].find_best_table_spec(table.sname, table.name)
+        spec = self.acl_specs["table_acls"].find_best_table_spec(table.schema.name, table.name)
         table.acls.clear()
         table.acl_bindings.clear()
         if spec is not None:
@@ -341,7 +347,7 @@ class AclConfig:
             self.set_node_acl_bindings(table, table, spec.get("acl_bindings"), spec.get("invalidate_bindings"))
         if self.verbose:
             print(
-                "set table {s}.{t} acls to {a}, bindings to {b}".format(s=table.sname, t=table.name, a=str(table.acls),
+                "set table {s}.{t} acls to {a}, bindings to {b}".format(s=table.schema.name, t=table.name, a=str(table.acls),
                                                                         b=str(table.acl_bindings)))
         for column in table.column_definitions:
             self.set_column_acls(column, table)
@@ -349,19 +355,19 @@ class AclConfig:
             self.set_fkey_acls(fkey, table)
 
     def set_column_acls(self, column, table):
-        spec = self.acl_specs["column_acls"].find_best_column_spec(column.sname, column.tname, column.name)
+        spec = self.acl_specs["column_acls"].find_best_column_spec(column.table.schema.name, column.table.name, column.name)
         column.acls.clear()
         column.acl_bindings.clear()
         if spec is not None:
             self.set_node_acl(column, spec)
             self.set_node_acl_bindings(column, table, spec.get("acl_bindings"), spec.get("invalidate_bindings"))
         if self.verbose:
-            print("set column {s}.{t}.{c} acls to {a}, bindings to {b}".format(s=column.sname, t=column.tname,
+            print("set column {s}.{t}.{c} acls to {a}, bindings to {b}".format(s=column.table.schema.name, t=column.table.name,
                                                                                c=column.name, a=str(column.acls),
                                                                                b=str(column.acl_bindings)))
 
     def set_fkey_acls(self, fkey, table):
-        spec = self.acl_specs["foreign_key_acls"].find_best_foreign_key_spec(fkey.sname, fkey.tname, fkey.names)
+        spec = self.acl_specs["foreign_key_acls"].find_best_foreign_key_spec(fkey.table.schema.name, fkey.table.name, fkey.names)
         fkey.acls.clear()
         fkey.acl_bindings.clear()
         if spec is not None:
@@ -397,17 +403,17 @@ class AclConfig:
             self.set_table_acls(table)
 
     def set_acls(self):
-        if isinstance(self.toplevel_config, ermrest_config.CatalogConfig):
+        if isinstance(self.toplevel_config, ermrest_model.Model):
             self.set_catalog_acls(self.toplevel_config)
-        elif isinstance(self.toplevel_config, ermrest_config.CatalogSchema):
+        elif isinstance(self.toplevel_config, ermrest_model.Schema):
             self.set_schema_acls(self.toplevel_config)
-        elif isinstance(self.toplevel_config, ermrest_config.CatalogTable):
+        elif isinstance(self.toplevel_config, ermrest_model.Table):
             self.set_table_acls(self.toplevel_config)
         else:
             raise ValueError("toplevel config is a {t}".format(t=str(type(self.toplevel_config))))
 
     def apply_acls(self):
-        self.toplevel_config.apply(self.catalog, self.saved_toplevel_config)
+        self.toplevel_config.apply(self.saved_toplevel_config)
 
     def dumps(self):
         """Dump a serialized (string) representation of the config.
@@ -521,7 +527,9 @@ class Table(AttrDict):
 class AclCLI(ConfigBaseCLI):
     def __init__(self):
         ConfigBaseCLI.__init__(self, "ACL configuration tool", None, version=VERSION)
-        self.parser.add_argument('-g', '--groups-only', help="create group table only", action="store_true")
+        group = self.parser.add_mutually_exclusive_group()
+        group.add_argument('-g', '--groups-only', help="create group table only", action="store_true")
+        group.add_argument('-o', '--omit-groups', help="do not create group table", action="store_true")
 
 
 def main():
@@ -530,7 +538,7 @@ def main():
     table_name = cli.get_table_arg(args)
     schema_names = cli.get_schema_arg_list(args)
     credentials = get_credential(args.host, args.credential_file)
-    save_groups = not args.dryrun
+    save_groups = not (args.dryrun or args.omit_groups)
     for schema in schema_names:
         acl_config = AclConfig(args.host, args.catalog, args.config_file, credentials, schema_name=schema,
                                table_name=table_name, verbose=args.verbose or args.debug)
