@@ -1,17 +1,26 @@
-
+import os
+import io
 import logging
 import datetime
+import codecs
+import csv
+import json
 
-from . import urlquote, datapath, DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE, Megabyte, get_transfer_summary
+from . import urlquote, datapath, DEFAULT_HEADERS, DEFAULT_CHUNK_SIZE, Megabyte, Kilobyte, get_transfer_summary, IS_PY2
 from .deriva_binding import DerivaBinding
 from .ermrest_config import CatalogConfig
 from . import ermrest_model
 
+
 class ErmrestCatalogMutationError(Exception):
     pass
 
+
 _clone_state_url = "tag:isrd.isi.edu,2018:clone-status"
 
+DEFAULT_PAGE_SIZE = 100000
+
+
 class ErmrestCatalog(DerivaBinding):
     """Persistent handle for an ERMrest catalog.
 
@@ -89,7 +98,7 @@ class ErmrestCatalog(DerivaBinding):
         return ermrest_model.Model.fromcatalog(self)
 
     def applyCatalogConfig(self, config):
-        return config.apply(self)
+        return config.apply()
 
     def getCatalogSchema(self):
         path = '/schema'
@@ -154,41 +163,171 @@ class ErmrestCatalog(DerivaBinding):
             return None
         return entity[0], entity[1]
 
-    def getAsFile(self, path, destfilename, headers=DEFAULT_HEADERS, callback=None):
+    def getAsFile(self,
+                  path,
+                  destfilename,
+                  headers=DEFAULT_HEADERS,
+                  callback=None,
+                  delete_if_empty=False,
+                  paged=False,
+                  page_size=DEFAULT_PAGE_SIZE):
         """
            Retrieve catalog data streamed to destination file.
            Caller is responsible to clean up file even on error, when the file may or may not be exist.
-
+           If "delete_if_empty" is True, the file will be inspected for "empty" content. In the case of
+           json/json-stream content, the presence of a single empty JSON object will be tested for. In the case of
+           CSV content, the file will be parsed with CSV reader to determine that only a single header line and no row
+           data is present.
         """
         self.check_path(path)
 
+        # Only entity API supported with paged mode at this time, otherwise fallback. We fallback rather than raise an
+        # exception in the case that the caller might be trying to perform an opportunistic paged request without
+        # knowing a priori if paged support for the given query is available.
+        page_size = page_size if page_size > 0 else DEFAULT_PAGE_SIZE
+        if not path.startswith("/entity") and paged:
+            logging.debug("Paged data retrieval only supported for entity API queries.")
+            paged = False
+
+        # Only "application/x-json-stream" or "text/csv" supported with paged mode at this time, otherwise fallback.
+        accept = headers.get("accept")
+        if accept not in ("application/x-json-stream", "text/csv"):
+            logging.debug("Paged data retrieval not supported for content type: %s" % accept)
+            paged = False
+
         headers = headers.copy()
 
         destfile = open(destfilename, 'w+b')
 
         try:
-            r = self._session.get(self._server_uri + path, headers=headers, stream=True)
-            self._response_raise_for_status(r)
-
             total = 0
             start = datetime.datetime.now()
-            logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
-            for buf in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
-                destfile.write(buf)
-                total += len(buf)
-                if callback:
-                    if not callback(progress="Downloading: %.2f MB transferred" % (float(total) / float(Megabyte))):
-                        destfile.close()
-                        return None
+            if not paged:
+                with self._session.get(self._server_uri + path, headers=headers, stream=True) as r:
+                    self._response_raise_for_status(r)
+                    content_type = r.headers.get("Content-Type")
+                    logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
+                    for buf in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
+                        destfile.write(buf)
+                        total += len(buf)
+                        if callback:
+                            if not callback(progress="Downloading: %.2f MB transferred" %
+                                                     (float(total) / float(Megabyte))):
+                                destfile.close()
+                                return
+                destfile.flush()
+                os.fsync(destfile.fileno())
+            else:
+                first_page = True
+                first_line = None
+                last_record = None
+                while True:
+                    qualifier = "@sort(RID)%s?limit=%d" % \
+                               (("@after(%s)" % urlquote(last_record)) if last_record is not None else "", page_size)
+                    # 1. Try to get a page worth of data, back-off page size if query run time errors are encountered
+                    with self._session.get(self._server_uri + path + qualifier, headers=headers) as r:
+                        if r.status_code == 400 and "Query run time limit exceeded" in r.text:
+                            r.close()
+                            page_size //= 2
+                            page_size = 1 if page_size < 1 else page_size
+                            logging.warning("Query runtime exceeded while attempting to transfer rows from %s to file "
+                                            "[%s]. The page size is being reduced to %s and the query will be retried."
+                                            % (self._server_uri + path, destfilename, page_size))
+                            continue
+
+                        # 2. Write the page to disk and determine the last RID processed in order to get the next page
+                        last_line = None
+                        content_type = r.headers.get("Content-Type")
+                        logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
+                        # CSV processing iterates over lines in the response, skipping the header line(s) in all but
+                        # the first page, and captures the last line of each page to determine the last record processed
+                        if content_type == "text/csv":
+                            skip = 1
+                            line_num = 0
+                            if first_page:
+                                lines = r.iter_lines(decode_unicode=True)
+                                reader = csv.reader(lines)
+                                first_line = next(reader)
+                                skip = reader.line_num
+                            for line in r.iter_lines():
+                                if not first_page:
+                                    line_num += 1
+                                    if line_num <= skip:
+                                        continue
+                                tline = line + b"\n"
+                                destfile.write(tline)
+                                total += len(tline)
+                                last_line = tline
+                            if last_line and last_line != first_line:
+                                reader = csv.DictReader([last_line.decode('utf-8')], first_line)
+                                last_line = next(reader)
+                            first_page = False
+                        # JSON-Stream processing writes the entire buffer to the destination file. The last line is
+                        # captured by reverse seeking in the buffer from right before the last b'\n' newline to the next
+                        # newline or buf[0], then calling readline from the current position
+                        elif content_type == "application/x-json-stream":
+                            buf = r.content
+                            if not buf:
+                                break
+                            destfile.write(buf)
+                            total += len(buf)
+                            b = io.BytesIO(buf)
+                            b.seek(-2, os.SEEK_END)
+                            while b.read(1) != b'\n':
+                                b.seek(-2, os.SEEK_CUR)
+                                if b.tell() == os.SEEK_SET:
+                                    break
+                            last_line = json.loads(b.readline().decode('utf-8'))
+
+                        # 3. Save the last record RID and flush the destination file buffers to disk.
+                        if not last_line:
+                            break
+                        destfile.flush()
+                        os.fsync(destfile.fileno())
+                        last_record = last_line['RID']
+                        if callback:
+                            if not callback(progress="Downloading: %.2f MB transferred" %
+                                                     (float(total) / float(Megabyte))):
+                                destfile.close()
+                                return
+
             elapsed = datetime.datetime.now() - start
             summary = get_transfer_summary(total, elapsed)
-            logging.info("File [%s] transfer successful. %s" % (destfilename, summary))
-            if callback:
-                callback(summary=summary, file_path=destfilename)
 
-            return r
+            # perform automatic file deletion on detected "empty" content, if requested
+            delete_file = True if total == 0 else False
+            if delete_if_empty and total > 0:
+                destfile.seek(0)
+                if content_type == "application/json" or content_type == "application/x-json-stream":
+                    buf = destfile.read(16)
+                    if buf == b"[]\n" or buf == b"{}\n":
+                        delete_file = True
+                elif content_type == "text/csv":
+                    reader = csv.reader(codecs.iterdecode(destfile, 'utf-8') if not IS_PY2 else destfile)
+                    rowcount = 0
+                    for row in reader:
+                        rowcount += 1
+                        if rowcount > 1:
+                            break
+                    if rowcount <= 1:
+                        delete_file = True
+
+            # automatically delete zero-length files or detected "empty" content
+            if delete_file:
+                destfile.close()
+                os.remove(destfilename)
+                destfile = None
+
+            log_msg = "File [%s] transfer successful. %s %s" % \
+                      (destfilename, summary,
+                       "File was automatically deleted due to empty content." if delete_file else "")
+            logging.info(log_msg)
+            if callback:
+                callback(summary=log_msg, file_path=destfilename)
+
         finally:
-            destfile.close()
+            if destfile:
+                destfile.close()
 
     def delete(self, path, headers=DEFAULT_HEADERS, guard_response=None):
         """Perform DELETE request, returning response object.
@@ -221,7 +360,13 @@ class ErmrestCatalog(DerivaBinding):
         else:
             raise ValueError('Catalog deletion refused when really is %s.' % really)
 
-    def clone_catalog(self, dst_catalog=None, copy_data=True, copy_annotations=True, copy_policy=True, truncate_after=True):
+    def clone_catalog(self,
+                      dst_catalog=None,
+                      copy_data=True,
+                      copy_annotations=True,
+                      copy_policy=True,
+                      truncate_after=True,
+                      exclude_schemas=None):
         """Clone this catalog's content into dest_catalog, creating a new catalog if needed.
 
         :param dst_catalog: Destination catalog or None to request creation of new destination (default).
@@ -229,6 +374,7 @@ class ErmrestCatalog(DerivaBinding):
         :param copy_annotations: Copy annotations when True (default).
         :param copy_policy: Copy access-control policies when True (default).
         :param truncate_after: Truncate destination history after cloning when True (default).
+        :param exclude_schemas: A list of schema names to exclude from the cloning process.
 
         When dest_catalog is provided, attempt an idempotent clone,
         assuming content MAY be partially cloned already using the
@@ -271,6 +417,7 @@ class ErmrestCatalog(DerivaBinding):
         new_keys = [] # ERMrest does not currently allow bulk key creation
         clone_states = {}
         fkeys_deferred = {}
+        exclude_schemas = [] if exclude_schemas is None else exclude_schemas
 
         def prune_parts(d):
             if not copy_annotations and 'annotations' in d:
@@ -335,6 +482,8 @@ class ErmrestCatalog(DerivaBinding):
             return (sname, tname, prune_parts(k.prejson()))
 
         for sname, schema in src_model.schemas.items():
+            if sname in exclude_schemas:
+                continue
             if sname not in dst_model.schemas:
                 new_model.append(copy_sdef(schema))
 
@@ -364,8 +513,8 @@ class ErmrestCatalog(DerivaBinding):
                         if cname not in src_columns:
                             raise ValueError("Destination column %s.%s.%s does not exist in source catalog." % (sname, tname, cname))
 
-                    src_keys = { tuple(sorted(key.unique_columns)): key for key in table.keys }
-                    dst_keys = { tuple(sorted(key.unique_columns)): key for key in dst_model.schemas[sname].tables[tname].keys }
+                    src_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in table.keys }
+                    dst_keys = { tuple(sorted(c.name for c in key.unique_columns)): key for key in dst_model.schemas[sname].tables[tname].keys }
 
                     for utuple in src_keys:
                         if utuple not in dst_keys:
