@@ -109,17 +109,19 @@ class HatracStore(DerivaBinding):
                 total = 0
                 start = datetime.datetime.now()
                 logging.debug("Transferring file %s to %s" % (self._server_uri + path, destfilename))
-                for buf in r.iter_content(chunk_size=Megabyte):
+                for buf in r.iter_content(chunk_size=DEFAULT_CHUNK_SIZE):
                     destfile.write(buf)
                     total += len(buf)
                     if callback:
                         if not callback(progress="Downloading: %.2f MB transferred" % (total / Megabyte)):
                             destfile.close()
+                            r.close()
                             os.remove(destfilename)
                             return None
-                destfile.flush()
                 elapsed = datetime.datetime.now() - start
                 summary = get_transfer_summary(total, elapsed)
+                destfile.flush()
+                os.fsync(destfile.fileno())
                 logging.info("File [%s] transfer successful. %s" % (destfilename, summary))
                 if callback:
                     callback(summary=summary, file_path=destfilename)
@@ -424,8 +426,10 @@ class HatracStore(DerivaBinding):
         url = resource_name + ';acl'
         if access:
             url += '/' + urlquote(access)
-        if role:
-            url += '/' + urlquote(role)
+            if role:
+                url += '/' + urlquote(role)
+        elif role:
+            raise ValueError('Do not specify "role" if "access" mode is not specified.')
         resp = self.get(url, headers)
         if resp.status_code == requests.codes.ok:
             if role:
@@ -444,7 +448,8 @@ class HatracStore(DerivaBinding):
         if 'add_role' is True, the operation will add a single role to the ACL, else it will attempt to replace
         all of the ACL's roles. This option is only valid when a list of one role is given.
         """
-        assert not add_role or len(roles) == 1, "Cannot 'add' more than one role at a time"
+        if add_role and len(roles) > 1:
+            raise ValueError("Cannot add more than one role at a time.")
         headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
         url = "%(resource_name)s;acl/%(access)s" % {'resource_name': resource_name, 'access': urlquote(access)}
         roles_obj = None
