@@ -9,7 +9,7 @@ import tempfile
 import logging
 import platform
 from collections import OrderedDict, namedtuple
-from deriva.core import ErmrestCatalog, CatalogConfig, HatracStore, HatracJobAborted, HatracJobPaused, \
+from deriva.core import ErmrestCatalog, HatracStore, HatracJobAborted, HatracJobPaused, \
     HatracJobTimeout, urlquote, stob, format_exception, get_credential, read_config, write_config, copy_config, \
     resource_path, make_dirs, lock_file, DEFAULT_CHUNK_SIZE, IS_PY2, __version__ as VERSION
 from deriva.core.utils import hash_utils as hu, mime_utils as mu, version_utils as vu
@@ -61,9 +61,10 @@ class DerivaUpload(object):
     DefaultTransferStateBaseName = ".deriva-upload-state"
     DefaultTransferStateFileName = "%s-%s.json"
 
-    def __init__(self, config_file=None, credential_file=None, server=None):
+    def __init__(self, config_file=None, credential_file=None, server=None, dcctx_cid=None):
         self.server_url = None
         self.catalog = None
+        self.catalog_model = None
         self.store = None
         self.config = None
         self.credentials = None
@@ -82,6 +83,7 @@ class DerivaUpload(object):
         self.override_config_file = config_file
         self.override_credential_file = credential_file
         self.server = self.getDefaultServer() if not server else server
+        self.dcctx_cid = dcctx_cid if dcctx_cid else self.__class__.__name__
         self.initialize()
 
     def __del__(self):
@@ -124,7 +126,7 @@ class DerivaUpload(object):
         self.store = HatracStore(protocol, host, self.credentials, session_config=session_config)
 
         # init dcctx cid to a default
-        self.set_dcctx_cid(self.__class__.__name__)
+        self.set_dcctx_cid(self.dcctx_cid)
 
         """
          Configuration initialization - this is a bit complex because we allow for:
@@ -180,6 +182,7 @@ class DerivaUpload(object):
         self.reset()
         self.config = None
         self.credentials = None
+        self.catalog_model = None
 
     def setServer(self, server):
         cleanup = self.server != server
@@ -192,6 +195,17 @@ class DerivaUpload(object):
         self.catalog.set_credentials(self.credentials, host)
         self.store.set_credentials(self.credentials, host)
 
+    def setConfig(self, config_file):
+        if not config_file:
+            config = self.getUpdatedConfig()
+            if config:
+                write_config(self.getDeployedConfigFilePath(), config)
+        else:
+            self._update_internal_config(read_config(config_file))
+        if not self.isVersionCompatible():
+            raise DerivaRestoreError("Upload version incompatibility detected",
+                                     "Current version: [%s], required version(s): %s." %
+                                     (self.getVersion(), self.getVersionCompatibility()))
     @classmethod
     def getDefaultServer(cls):
         servers = cls.getServers()
@@ -331,7 +345,7 @@ class DerivaUpload(object):
                (self.DefaultTransferStateBaseName, self.server.get('host', 'localhost'))
 
     def getRemoteConfig(self):
-        catalog_config = CatalogConfig.fromcatalog(self.catalog)
+        catalog_config = self.catalog.getCatalogModel()
         return catalog_config.bulk_upload
 
     def getUpdatedConfig(self):
@@ -470,6 +484,7 @@ class DerivaUpload(object):
         return None, None, None
 
     def uploadFiles(self, status_callback=None, file_callback=None):
+        self.catalog_model = self.catalog.getCatalogModel()
         for group, assets in self.file_list.items():
             for asset_group_num, asset_mapping, groupdict, file_path in assets:
                 if self.cancelled:
@@ -480,7 +495,11 @@ class DerivaUpload(object):
                     if status_callback:
                         status_callback()
                     self.uploadFile(file_path, asset_mapping, groupdict, file_callback)
-                    self.file_status[file_path] = FileUploadState(UploadState.Success, "Complete")._asdict()
+                    if self.cancelled:
+                        self.file_status[file_path] = FileUploadState(UploadState.Cancelled,
+                                                                      "Cancelled by user")._asdict()
+                    else:
+                        self.file_status[file_path] = FileUploadState(UploadState.Success, "Complete")._asdict()
                 except HatracJobPaused:
                     status = self.getTransferStateStatus(file_path)
                     if status:
@@ -546,6 +565,8 @@ class DerivaUpload(object):
             alg = alg.lower()
             self.metadata[alg] = checksum[0]
             self.metadata[alg + "_base64"] = checksum[1]
+        if self.cancelled:
+            return
 
         # 3. Execute any configured preprocessors
         self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=PRE_PROCESSORS_KEY)
@@ -589,8 +610,14 @@ class DerivaUpload(object):
         column_map = asset_mapping.get("column_map", {})
         updated_record = self.interpolateDict(self.metadata, column_map)
         if updated_record != record:
+            record_update_template = asset_mapping.get("record_update_template")
+            require_record_update_template = stob(asset_mapping.get("require_record_update_template", False))
+            if require_record_update_template and not record_update_template:
+                raise DerivaUploadCatalogUpdateError(
+                    "A required 'record_update_template' parameter for this asset mapping could not be found in the "
+                    "configuration. The record will not be updated.")
             logger.info("Updating catalog for file [%s]" % self.getFileDisplayName(file_path))
-            self._catalogRecordUpdate(self.metadata['target_table'], record, updated_record)
+            self._catalogRecordUpdate(self.metadata['target_table'], record, updated_record, record_update_template)
 
         # 9. Execute any configured post_processors
         self._execute_processors(file_path, asset_mapping, match_groupdict, processor_list=POST_PROCESSORS_KEY)
@@ -637,7 +664,7 @@ class DerivaUpload(object):
             raise DerivaUploadConfigurationError("Record query template substitution error: %s" % format_exception(e))
         result = self.catalog.get(path).json()
         if result:
-            self._updateFileMetadata(result[0])
+            self._updateFileMetadata(result[0], no_overwrite=True)
             return self.pruneDict(result[0], column_map)
         else:
             row = self.interpolateDict(self.metadata, column_map)
@@ -666,16 +693,22 @@ class DerivaUpload(object):
 
         self._urlEncodeMetadata(asset_mapping.get("url_encoding_safe_overrides"))
 
-    def _updateFileMetadata(self, src, strict=False):
+    def _updateFileMetadata(self, src, strict=False, no_overwrite=False):
         if not (isinstance(src, dict)):
             ValueError("Invalid input parameter type(s): (src = %s), expected (dict)" % type(src).__name__)
-        if strict:
-            for k in src.keys():
+        dst = src.copy()
+        for k in src.keys():
+            if strict:
                 if k in UploadMetadataReservedKeyNames:
                     logger.warning("Context metadata update specified reserved key name [%s], "
                                    "ignoring value: %s " % (k, src[k]))
-                    del src[k]
-        self.metadata.update(src)
+                    del dst[k]
+                    continue
+            # dont overwrite any existing metadata field
+            if no_overwrite:
+                if k in self.metadata:
+                    del dst[k]
+        self.metadata.update(dst)
 
     def _queryFileMetadata(self, asset_mapping):
         """
@@ -785,6 +818,16 @@ class DerivaUpload(object):
     def _validate_catalog_row_columns(self, row, table):
         return set(row.keys()) - self._get_catalog_table_columns(table)
 
+    def _validate_row_key_constraints(self, catalog_table, row):
+        schema_name, table_name = self.catalog.splitQualifiedCatalogName(catalog_table)
+        schema = self.catalog_model.schemas.get(schema_name)
+        table = schema.tables.get(table_name)
+        non_null_correlations = {cname for cname, cval in row.items() if cval is not None}
+        for key in table.keys:
+            if set(key.unique_columns).issubset(non_null_correlations):
+                return True  # it is safe
+        return False  # it is not safe
+
     def _get_catalog_default_columns(self, row, table, exclude=None, quote_url=True):
         columns = self._get_catalog_table_columns(table)
         if isinstance(exclude, list):
@@ -829,7 +872,7 @@ class DerivaUpload(object):
             (etype, value, traceback) = sys.exc_info()
             raise DerivaUploadCatalogCreateError(format_exception(value))
 
-    def _catalogRecordUpdate(self, catalog_table, old_row, new_row):
+    def _catalogRecordUpdate(self, catalog_table, old_row, new_row, record_update_template=None):
         """
 
         :param catalog_table:
@@ -847,22 +890,30 @@ class DerivaUpload(object):
                 raise RuntimeError("Cannot update catalog - "
                                    "new row column list and old row column list do not match: New: %s != Old: %s" %
                                    (keys, old_keys))
-            combined_row = {
+            o_keys = ','.join(["o%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))])
+            n_keys = ','.join(["n%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))])
+            update_row = {
                 'o%d' % i: old_row[keys[i]]
                 for i in range(len(keys))
             }
-            combined_row.update({
+            update_row.update({
                 'n%d' % i: new_row[keys[i]]
                 for i in range(len(keys))
             })
-            update_uri = '/attributegroup/%s/%s;%s' % (
-                catalog_table,
-                ','.join(["o%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))]),
-                ','.join(["n%d:=%s" % (i, urlquote(keys[i])) for i in range(len(keys))])
-            )
+            if record_update_template:
+                update_uri = record_update_template.format(**self.metadata)
+                update_row = new_row
+            else:
+                update_uri = '/attributegroup/%s/%s;%s' % (catalog_table, o_keys, n_keys)
+                if self.config.get("strict_update_check", True) and not \
+                        self._validate_row_key_constraints(catalog_table, old_row):
+                    raise ValueError(
+                        "Potential unsafe attributegroup update [%s]: at least one pre-existing, non-null correlation "
+                        "key is required. Old values: %s, New values: %s" %
+                        (update_uri, json.dumps(old_row), json.dumps(new_row)))
             logger.debug(
-                "Attempting catalog record update [%s] with data: %s" % (update_uri, json.dumps(combined_row)))
-            return self.catalog.put(update_uri, json=[combined_row]).json()
+                "Attempting catalog record update [%s] with data: %s" % (update_uri, json.dumps(update_row)))
+            return self.catalog.put(update_uri, json=[update_row]).json()
         except:
             (etype, value, traceback) = sys.exc_info()
             raise DerivaUploadCatalogUpdateError(format_exception(value))
@@ -1013,6 +1064,7 @@ class DerivaUpload(object):
             self.transfer_state_fh.truncate()
             json.dump(self.transfer_state, self.transfer_state_fh, indent=2)
             self.transfer_state_fh.flush()
+            os.fsync(self.transfer_state_fh.fileno())
         except Exception as e:
             logger.warning("Unable to write transfer state file: %s" % format_exception(e))
 
@@ -1020,6 +1072,7 @@ class DerivaUpload(object):
         if self.transfer_state_fh and not self.transfer_state_fh.closed:
             try:
                 self.transfer_state_fh.flush()
+                os.fsync(self.transfer_state_fh.fileno())
                 self.transfer_state_fh.close()
             except Exception as e:
                 logger.warning("Unable to flush/close transfer state file: %s" % format_exception(e))
@@ -1040,8 +1093,12 @@ class DerivaUpload(object):
 
 class GenericUploader(DerivaUpload):
 
-    def __init__(self, config_file=None, credential_file=None, server=None):
-        DerivaUpload.__init__(self, config_file, credential_file, server)
+    def __init__(self, config_file=None, credential_file=None, server=None, dcctx_cid=None):
+        DerivaUpload.__init__(self,
+                              config_file=config_file,
+                              credential_file=credential_file,
+                              server=server,
+                              dcctx_cid=dcctx_cid)
 
     @classmethod
     def getVersion(cls):
